from math import *
from geometry3d_new import *
from openscadWriter import *
import sys

# Create an stl file for pieces to Stewart Coffin's Alphabet Stew puzzle
# Each piece is divided into components, the basic component
# uses 8 vertices from the 'vertices' with faces mpf.
# The vector 'comps' has the vertex set for the basic component rotated.
# Each piece is then constructed from 2-5 components, as given in 'mypiece'.
# Faces connecting the components are duplicates and removed.
# The internal faces marked by 'shaveface' are then shaved according to the offset.
# The resulting piece is output as an openSCAD program.
# To get the stl file, open this in openSCAD, then render and export it.

# Four component pieces are named A-I or X.
# Pieces with n components are labeled Pn.m, where m is the piece number

# To use:
# 1) Ensure that geometry3d_new.py and openscadWriter.py are present.
# 2) Set the size, truncation, shaving and offset of the pieces
#    (f2fsize, tr, shave and offset below).
# 3) Run this python program to generate an openSCAD file.
# 4) Open the openSCAD file to render the piece and export to stl.

# Syntax:
# python stewPuzzle2.py piecename
# example: python stewPuzzle2.py D
# example: python stewPuzzle2.py P5.12d

# Determine if we have a duplicate face so we can remove it.
# We can also detect when two components are connected.
def facesMatch(f1,f2):
  if len(f1)!=len(f2): return 0
  l = len(f1)

  matchf = 0
  matchb = 0
  for i in range(0,l):
    for j in range(0,l):
      if f1[i]==f2[j]:
        matchf = 1
        matchb = 1
        for k in range(1,l):
          if f1[(i+k)%l] != f2[(j+k)%l]: matchf = 0
          if f1[(i+k)%l] != f2[(j-k)%l]: matchb = 0
        if matchf>0: return 1
        if matchb>0: return -1
  return 0

myname = "stewPuzzle2.py"
# Set the size of the puzzle and other parameters ...
f2fsize = 64.0 # For the wood version sent by Stewart Coffin
tr = 0.4 # Determines inner trucation level, taken from wood pieces
#tr = 0.0 # No truncation
shave = 1 # Set this to 1 to shave the internal faces
offset = 0.1 # Amount in mm to shave off faces
#offset = 0.5 # Amount in mm to shave off faces

# The piece to make
piece = " " # Default
if len(sys.argv)>1:
  piece = sys.argv[1]

# Check to be sure we can make this piece
test = ord(piece[0:1])
if (test<65 or test>73) and test!=48 and test!=80 and test!=88:
  print("I don't know how to make piece "+piece)
  print("Syntax to make piece D:")
  print("python",myname,"D")
  exit(1)

# Calculated parameters
scale = 0.5*f2fsize/sqrt(2.0) # Main scale factor when main RD vertices are (2,0,0)
edgelength = sqrt(3.0)*scale # Edge length of untruncated RD
diameter = 4.0*scale # RD size (untruncated)
print("Scale is ",scale)

# Vertices needed to make a Hollow Garnet
# 6 RD Vertices (truncated)
# 8 more RD vertices
# 24 half vertices
# 6 inner RD vertices (not truncated)
# 8 more inner vertices
# 24 inner half vertices
# Total 76 vertices
vertices = [[0,0,6],[6,0,0],[0,6,0],[-6,0,0],[0,-6,0],[0,0,-6], # 0-5 (RD vertices)
            [4,4,4],[-4,4,4],[-4,-4,4],[4,-4,4],                # 6-9
            [4,4,-4],[-4,4,-4],[-4,-4,-4],[4,-4,-4],            # 10-13
            [2,2,6],[-2,2,6],[-2,-2,6],[2,-2,6],                # 14-17 (half vertices)
            [6,2,2],[2,6,2],[-2,6,2],[-6,2,2],                  # 18-21
            [-6,-2,2],[-2,-6,2],[2,-6,2],[6,-2,2],              # 22-25
            [6,2,-2],[2,6,-2],[-2,6,-2],[-6,2,-2],              # 26-29
            [-6,-2,-2],[-2,-6,-2],[2,-6,-2],[6,-2,-2],          # 30-33
            [2,2,-6],[-2,2,-6],[-2,-2,-6],[2,-2,-6],            # 34-37
            [0,0,4],[4,0,0],[0,4,0],[-4,0,0],[0,-4,0],[0,0,-4], # 38-43 (inner RD vertices)
            [2,2,2],[-2,2,2],[-2,-2,2],[2,-2,2],                # 44-47
            [2,2,-2],[-2,2,-2],[-2,-2,-2],[2,-2,-2],            # 48-51
            [1,1,3],[-1,1,3],[-1,-1,3],[1,-1,3],                # 52-55 (inner half vertices)
            [3,1,1],[1,3,1],[-1,3,1],[-3,1,1],                  # 56-59
            [-3,-1,1],[-1,-3,1],[1,-3,1],[3,-1,1],              # 60-63
            [3,1,-1],[1,3,-1],[-1,3,-1],[-3,1,-1],              # 64-67
            [-3,-1,-1],[-1,-3,-1],[1,-3,-1],[3,-1,-1],          # 68-71
            [1,1,-3],[-1,1,-3],[-1,-1,-3],[1,-1,-3]];           # 72-75

# Scale the vertices as requested by the user
inner_vertices = 37
end_vertices = len(vertices)
for i in range(0,inner_vertices+1):
  for j in range(0,3):
    vertices[i][j] = scale*0.25*vertices[i][j]
for i in range(inner_vertices+1,end_vertices):
  for j in range(0,3):
    vertices[i][j] = scale*0.5*tr*vertices[i][j]

#print("end_vertices is ",end_vertices)
#print(vertices)

# main part faces on vertices 0-9
# The first 2 faces are external, the 3rd is inside,
# and the remaining 5 faces are internal and shaved.
mpf = [[0,1,4], [1,2,3,4], [9,8,7,6,5],
      [1,0,5,6], [2,1,6,7], [3,2,7,8], [4,3,8,9], [0,4,9,5]]
compfaces = len(mpf)
            
# Each piece is made from 2-5 components, each here is rotated differently
comps = [[0, 15, 19, 6, 14, 38, 53, 57, 44, 52],  # Main piece vertices
       [0, 14, 25, 9, 17, 38, 52, 63, 47, 55],  # -90 degree rotation in z (vertices)
       [0, 16, 21, 7, 15, 38, 54, 59, 45, 53],  # +90 in z
       [0, 17, 23, 8, 16, 38, 55, 61, 46, 54],  # +180 in z
       [4, 23, 17, 9, 24, 42, 61, 55, 47, 62],  # +90 in x
       [3, 21, 16, 8, 22, 41, 59, 54, 46, 60],  # R05
       [4, 24, 33, 13, 32, 42, 62, 71, 51, 70], # R06
       [4, 31, 22, 8, 23, 42, 69, 60, 46, 61],  # R07
       [2, 27, 18, 6, 19, 40, 65, 56, 44, 57],  # R08
       [1, 33, 24, 9, 25, 39, 71, 62, 47, 63],  # R09
       [1, 26, 37, 13, 33, 39, 64, 75, 51, 71], # R10
       [1, 25, 14, 6, 18, 39, 63, 52, 44, 56],  # R11
       [2, 27, 18, 6, 19, 40, 65, 56, 44, 57],  # R12 == R08
       [2, 28, 34, 10, 27, 40, 66, 72, 48, 65]] # R13

# Piece definitions in terms of components
mypiece = []

# 1 Component
if piece=="0" or piece=="P1.1":
  mypiece = [0]

# 4 Components
if piece=="A" or piece=="P4.1":
  mypiece = [0,1,3,5]
if piece=="B" or piece=="P4.2":
  mypiece = [0,1,4,6]
if piece=="C" or piece=="P4.3":
  mypiece = [0,2,11,8]
if piece=="D" or piece=="P4.4":
  mypiece = [0,1,3,7]
if piece=="E" or piece=="P4.5":
  mypiece = [0,1,3,8]
if piece=="F" or piece=="P4.6":
  mypiece = [0,1,3,9]
if piece=="G" or piece=="P4.7":
  mypiece = [0,1,9,10]
if piece=="H" or piece=="P4.8":
  mypiece = [0,1,9,8]
if piece=="I" or piece=="P4.9":
  mypiece = [0,1,8,4]
if piece=="X" or piece=="P4.10":
  mypiece = [0,1,2,3]

# 2 Components
if piece=="P2.1":
  mypiece = [0,1]
if piece=="P2.2":
  mypiece = [0,8]

# 3 Components
if piece=="P3.1":
  mypiece = [0,11,8]
if piece=="P3.2":
  mypiece = [0,1,8]
if piece=="P3.3":
  mypiece = [0,1,9]
if piece=="P3.4":
  mypiece = [0,1,2]
if piece=="P3.5":
  mypiece = [0,1,4]

# 5 Components
# Needed by "Knife attack"
if piece=="P5.4":
  mypiece = [0,1,3,5,8]
if piece=="P5.5":
  mypiece = [0,1,3,5,7]
if piece=="P5.7":
  mypiece = [0,1,3,8,13]
if piece=="P5.11":
  mypiece = [0,2,11,8,9]
if piece=="P5.15":
  mypiece = [0,1,2,3,7]
if piece=="P5.18":
  mypiece = [0,1,4,8,7]
if piece=="P5.12d":
  mypiece = [0,1,3,8,11]

if (len(mypiece)==0):
  print("I don't know how to make piece "+piece)
  print("Syntax to make piece D:")
  print("python",myname,"D")
  exit(1)

# Make the piece
myfaces = []
shaveface = []
for i in range(0,len(mypiece)):
  for j in range(0,len(mpf)):
    newface = []
    for k in range(0,len(mpf[j])):
      newface.append(comps[mypiece[i]][mpf[j][k]])
    myfaces.append(newface)
    if j>2: shaveface.append(1)
    else: shaveface.append(0)

# print(myfaces)
# print(shaveface)

# On the main piece, we now remove duplicate faces

markForDelete = []
for i in range(0,len(myfaces)): markForDelete.append(-1)

for i in range(0,len(myfaces)):
  for j in range(i+1,len(myfaces)):
    test = facesMatch(myfaces[i],myfaces[j])
    if test<0:
      markForDelete[i] = int(j/compfaces) # This gives the component number this face connects to
      markForDelete[j] = int(i/compfaces)
    if test>0:
      print("Face matches, but not reversed??",i,j)

#print(markForDelete)

# Now remove the duplicate faces
facesRemoved = 0
totalFaces=len(myfaces)
for i in range(len(myfaces)-1,-1,-1):
  if markForDelete[i]>=0:
    facesRemoved = facesRemoved+1
    del myfaces[i]
    del shaveface[i]
print("Faces: ",facesRemoved," out of ",totalFaces," were removed, now left with ",len(myfaces)," faces.")

if shave==0:
  pieceFile = "piece" + piece + ".scad"
else:
  pieceFile = "piece" + piece + "_" +str(offset) + ".scad"

fo = open(pieceFile,"w")

fo.write('scale = '+str(scale)+';\n')
fo.write('// RD face to face distance = '+str(f2fsize)+' mm\n')
fo.write('// edgelength = '+str(edgelength)+' mm\n')
fo.write('// diameter = '+str(diameter)+' mm\n')
fo.write('// inner RD face to face distance = '+str(f2fsize*tr)+' mm\n')
if shave>0:
  fo.write('// offset = '+str(offset)+'\n')

fo.write('rotate(a=[0,180,0])\n')
fo.write('difference() {\n')
writePolyhedronSpecial(fo, vertices, myfaces, 0, 1)

# Note: only faces with 3 or 4 edges will be shaved
if shave>0:
  shFaces = []
  notShaved = []
  for i in range(0,len(myfaces)):
    if shaveface[i]>0:
      if len(myfaces[i])==3:
        shFaces.append([myfaces[i][0], myfaces[i][1], myfaces[i][2]])
      if len(myfaces[i])==4:
        shFaces.append([myfaces[i][0], myfaces[i][1], myfaces[i][2], myfaces[i][3]])

  fo.write('// Shaved faces\n')
  for i in range(0,len(shFaces)):
    shaveFace(fo, vertices, shFaces, i, offset)

fo.write('} // end difference\n')
fo.close

print("Done writing openSCAD file: ",pieceFile)

